# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: connector.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'connector.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x63onnector.proto\x12\x12\x63hat_bot.connector\"%\n\x14\x43onnectStreamRequest\x12\r\n\x05\x66rame\x18\x01 \x01(\x0c\"&\n\x15\x43onnectStreamResponse\x12\r\n\x05\x66rame\x18\x01 \x01(\x0c\x32w\n\tConnector\x12j\n\rConnectStream\x12(.chat_bot.connector.ConnectStreamRequest\x1a).chat_bot.connector.ConnectStreamResponse\"\x00(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'connector_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_CONNECTSTREAMREQUEST']._serialized_start=39
  _globals['_CONNECTSTREAMREQUEST']._serialized_end=76
  _globals['_CONNECTSTREAMRESPONSE']._serialized_start=78
  _globals['_CONNECTSTREAMRESPONSE']._serialized_end=116
  _globals['_CONNECTOR']._serialized_start=118
  _globals['_CONNECTOR']._serialized_end=237
# @@protoc_insertion_point(module_scope)
